
//******************************************************************************
// RCF - Remote Call Framework
//
// Copyright (c) 2005 - 2011, Delta V Software. All rights reserved.
// http://www.deltavsoft.com
//
// RCF is distributed under dual licenses - closed source or GPL.
// Consult your particular license for conditions of use.
//
// Version: 1.3.1
// Contact: support <at> deltavsoft.com 
//
//******************************************************************************

// Copyright (C) 2001-2003
// William E. Kempf
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  William E. Kempf makes no representations
// about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.

#ifndef BOOST_THREAD_NO_TSS_CLEANUP

namespace boost {

typedef ::std::vector<void*> tss_slots;

struct tss_data_t
{
    mutex mutex_;
    ::std::vector< ::boost::function1<void, void*>*> cleanup_handlers;
#if defined(BOOST_HAS_WINTHREADS)
    DWORD native_key;
#elif defined(BOOST_HAS_PTHREADS)
    pthread_key_t native_key;
#elif defined(BOOST_HAS_MPTASKS)
    TaskStorageIndex native_key;
#endif
};

tss_data_t* tss_data = 0;
once_flag tss_data_once = RCF_BOOST_ONCE_INIT;

extern "C" void cleanup_slots(void* p)
{
    tss_slots* tlSlots = static_cast<tss_slots*>(p);
    for (tss_slots::size_type i = 0; i < tlSlots->size(); ++i)
    {
        mutex::scoped_lock lock(tss_data->mutex_);
        (*tss_data->cleanup_handlers[i])((*tlSlots)[i]);
        (*tlSlots)[i] = 0;
    }
}

void init_tss_data()
{
    ::std::auto_ptr<tss_data_t> temp(new tss_data_t);

#if defined(BOOST_HAS_WINTHREADS)
    //Force the cleanup implementation library to be linked in
    RBT_tss_cleanup_implemented();

    //Allocate tls slot
    temp->native_key = TlsAlloc();
    if (temp->native_key == 0xFFFFFFFF)
        return;
#elif defined(BOOST_HAS_PTHREADS)
    int res = pthread_key_create(&temp->native_key, &cleanup_slots);
    if (res != 0)
        return;
#elif defined(BOOST_HAS_MPTASKS)
    OSStatus status = MPAllocateTaskStorageIndex(&temp->native_key);
    if (status != noErr)
        return;
#endif

    // Intentional memory "leak"
    // This is the only way to ensure the mutex in the global data
    // structure is available when cleanup handlers are run, since the
    // execution order of cleanup handlers is unspecified on any platform
    // with regards to C++ destructor ordering rules.
    tss_data = temp.release();
}

#if defined(BOOST_HAS_WINTHREADS)
tss_slots* get_slots(bool alloc);

void __cdecl tss_thread_exit()
{
    tss_slots* tlSlots = get_slots(false);
    if (tlSlots)
        cleanup_slots(tlSlots);
}
#endif

tss_slots* get_slots(bool alloc)
{
    tss_slots* tlSlots = 0;

#if defined(BOOST_HAS_WINTHREADS)
    tlSlots = static_cast<tss_slots*>(
        TlsGetValue(tss_data->native_key));
#elif defined(BOOST_HAS_PTHREADS)
    tlSlots = static_cast<tss_slots*>(
        pthread_getspecific(tss_data->native_key));
#elif defined(BOOST_HAS_MPTASKS)
    tlSlots = static_cast<tss_slots*>(
        MPGetTaskStorageValue(tss_data->native_key));
#endif

    if (tlSlots == 0 && alloc)
    {
        ::std::auto_ptr<tss_slots> temp(new tss_slots);

#if defined(BOOST_HAS_WINTHREADS)
        if (RBT_at_thread_exit(&tss_thread_exit) == -1)
            return 0;
        if (!TlsSetValue(tss_data->native_key, temp.get()))
            return 0;
#elif defined(BOOST_HAS_PTHREADS)
        if (pthread_setspecific(tss_data->native_key, temp.get()) != 0)
            return 0;
#elif defined(BOOST_HAS_MPTASKS)
        if (MPSetTaskStorageValue(tss_data->native_key, temp.get()) != noErr)
            return 0;
#endif

        tlSlots = temp.release();
    }

    return tlSlots;
}

} // namespace boost

namespace boost {

namespace detail {
    void tss::init(::boost::function1<void, void*>* pcleanup)
{
    boost::call_once(&init_tss_data, tss_data_once);
    if (tss_data == 0)
        throw thread_resource_error();
    mutex::scoped_lock lock(tss_data->mutex_);
    try
    {
        tss_data->cleanup_handlers.push_back(pcleanup);
        m_slot = tss_data->cleanup_handlers.size() - 1;
    }
    catch (...)
    {
        throw thread_resource_error();
    }
}

void* tss::get() const
{
    tss_slots* tlSlots = get_slots(false);

    if (!tlSlots)
        return 0;

    if (m_slot >= tlSlots->size())
        return 0;

    return (*tlSlots)[m_slot];
}

void tss::set(void* value)
{
    tss_slots* tlSlots = get_slots(true);

    if (!tlSlots)
        throw thread_resource_error();

    if (m_slot >= tlSlots->size())
    {
        try
        {
            tlSlots->resize(m_slot + 1);
        }
        catch (...)
        {
            throw thread_resource_error();
        }
    }

    (*tlSlots)[m_slot] = value;
}

void tss::cleanup(void* value)
{
    mutex::scoped_lock lock(tss_data->mutex_);
    (*tss_data->cleanup_handlers[m_slot])(value);
}

} // namespace detail
} // namespace boost

#endif //BOOST_THREAD_NO_TSS_CLEANUP
